@@ -1,5 +1,6 @@
 import os
 from pathlib import Path
+import numpy
 import pandas as pd
 import github
 import pickle
@@ -27,7 +28,7 @@ class Utility():
             Save a data list to a pandas table.
         get_repo_informations(data_root_dir)
             Get a repository data (owner and name).
-        get_repos(repo_name_pattern, blacklist_pattern, token, data_root_dir)
+        get_repos(token, data_root_dir, whitelist_patterns=None, blacklist_patterns=None)
             Get mutiple repositorys by pattern and token.
         get_repo(repo_owner, repo_name, token, data_root_dir)
             Get a repository by owner, name and token.
@@ -53,6 +54,8 @@ class Utility():
             Extracting general event data from a issue or pull request.
         extract_comment_data(comment, parent_id, parent_name, users_ids, data_root_dir)
             Extracting general comment data from a pull request or issue.
+        define_unknown_user(user_dict, unknown_user, data_root_dir)
+            Defines a unknown user. Add unknown user to alias.
     
     """
     USERS = "Users.p"
@@ -174,22 +177,22 @@ class Utility():
         return None, None
     
     @staticmethod
-    def get_repos(whitelist_pattern, blacklist_pattern, token, data_root_dir):
+    def get_repos(token, data_root_dir, whitelist_patterns=None, blacklist_patterns=None):
         """
-        get_repos(repo_name_pattern, blacklist_pattern, token, data_root_dir)
+        get_repos(token, data_root_dir, whitelist_patterns=None, blacklist_patterns=None)
 
-        Get mutiple repositorys by pattern and token.
+        Get mutiple repositorys by mutiple pattern and token.
 
         Parameters
         ----------
-        whitelist_pattern : str
-            the whitelist pattern of the desired repository.
-        blacklist_pattern : str
-            the blacklist pattern of the desired repository.
         token : str
             A valid Github Token.
         data_root_dir : str
             Data root directory for the repositorys.
+        whitelist_patterns : list
+            the whitelist pattern of the desired repository.
+        blacklist_patterns : list
+            the blacklist pattern of the desired repository.
         
         Returns
         -------
@@ -201,11 +204,26 @@ class Utility():
             PyGithub Repository object structure: https://pygithub.readthedocs.io/en/latest/github_objects/Repository.html
 
         """
+
         g = github.Github(token)
         relevant_repos = []
         for repo in g.get_user().get_repos():
-            if whitelist_pattern == "" or whitelist_pattern == None or whitelist_pattern in repo.name:
-                if blacklist_pattern == "" or blacklist_pattern == None or not blacklist_pattern in repo.name:
+            whitelist_pass = False
+            if whitelist_patterns == [] or whitelist_patterns == None:
+                whitelist_pass = True
+            else:
+                for whitelist_pattern in whitelist_patterns:
+                    if whitelist_pattern in repo.name:
+                        whitelist_pass = True
+                        break
+            if whitelist_pass:
+                blacklist_pass = True
+                if blacklist_patterns != [] or blacklist_patterns is not None:
+                    for blacklist_pattern in blacklist_patterns:
+                        if blacklist_pattern in repo.name:
+                            blacklist_pass = False
+                            break
+                if blacklist_pass:
                     repo_dir = Path(data_root_dir, repo.owner.login + "/" + repo.name)
                     repo_dir.mkdir(parents=True, exist_ok=True)
                     repo_file = Path(repo_dir, Utility.REPO)
@@ -402,6 +420,7 @@ class Utility():
             Dict of User Ids as Keys and anonym Ids as Value.
         data_root_dir : str
             Repo dir of the project.
+        
         Returns
         -------
         str
@@ -413,8 +432,7 @@ class Utility():
 
         """
         if not user:
-            print("No User. This call was slow!!! Should be fixed")
-            return
+            return None
         if user.node_id in users_ids:
             return users_ids[user.node_id]
         users_file = Path(data_root_dir, Utility.USERS)
@@ -423,14 +441,27 @@ class Utility():
             users_df = pd.read_pickle(users_file)
         user_data = {}
         user_data["anonym_uuid"] = generate_id(seed=user.node_id)
-        users_ids[user.node_id] = user_data["anonym_uuid"]
         user_data["id"] = user.node_id
         try:
             user_data["name"] = user.name
         except:
-            user_data["name"] = "unknown User"
-        user_data["email"] = user.email
-        user_data["login"] = user.login
+            print("No User name in:")
+            print(data_root_dir)
+        try:
+            user_data["email"] = user.email
+        except:
+            print("No User email in:")
+            print(data_root_dir)
+        try:
+            user_data["login"] = user.login
+        except:
+            print("No User login in:")
+            print(data_root_dir)
+        if "login" in user_data:
+            if user_data["login"] == "invalid-email-address" and not "name" in user_data:
+                return None
+
+        users_ids[user.node_id] = user_data["anonym_uuid"]
         users_df = users_df.append(user_data, ignore_index=True)
         with open(users_file, "wb") as f:
             pickle.dump(users_df, f)
@@ -469,6 +500,8 @@ class Utility():
         commit = repo.get_commit(sha)
         if not commit:
             return None
+        if commit._author == github.GithubObject.NotSet:
+            return None
         return Utility.extract_user_data(commit.author, users_ids, data_root_dir)
                
     @staticmethod
@@ -504,6 +537,8 @@ class Utility():
         commit = repo.get_commit(sha)
         if not commit:
             return None
+        if commit._committer == github.GithubObject.NotSet:
+            return None
         return Utility.extract_user_data(commit.committer, users_ids, data_root_dir)
 
     @staticmethod
@@ -631,3 +666,57 @@ class Utility():
             comment_data["author"] = Utility.extract_user_data(comment.user, users_ids, data_root_dir)
         return comment_data
     
+    @staticmethod
+    def define_unknown_user(user_dict, unknown_user, data_root_dir):
+        """
+        define_unknown_user(user_dict, unknown_user, data_root_dir)
+
+        Defines a unknown user. Add unknown user to alias.
+
+        Parameters
+        ----------
+        user_dict: dict
+            Dictionary which contains users. 
+        unknown_user : str
+            Name of a unknown user.
+        data_root_dir : str
+            Repo dir of the project.
+
+        Returns
+        -------
+        str
+            Uuid of the user.
+
+        Notes
+        -----
+            Example User Dict: {"unknown_user": "real user node id"}
+            If the real user node id does not exist in the users table then a new user will be created
+
+        """
+        users = Utility.get_users(data_root_dir)
+        if unknown_user in user_dict:
+            p_user = users.loc[users.id == user_dict[unknown_user]]
+            if not p_user.empty:
+                alias = ""
+                user = p_user.iloc[0]
+                if "alias" in user:
+                    if (user["alias"] == numpy.NAN) or (user["alias"] is None):
+                        alias = unknown_user
+                    else:
+                        alias = user["alias"] + ";" + unknown_user
+                else:
+                    alias = unknown_user
+                users.loc[users.id == user_dict[unknown_user], 'alias'] = alias
+                pd_file = Path(data_root_dir, Utility.USERS)
+                with open(pd_file, "wb") as f:
+                    pickle.dump(users, f)
+                return user["anonym_uuid"]
+            
+            class UserData:
+                node_id = user_dict[unknown_user]
+                name = numpy.NaN
+                email = numpy.NaN
+                login = numpy.NaN
+                alias = unknown_user  
+            users_ids = Utility.get_users_ids(data_root_dir)
+            return Utility.extract_user_data(UserData(),users_ids,data_root_dir)
\ No newline at end of file
